import json

import requests
import time

from metasdk.logger import LOGGER_ENTITY

from metasdk.exceptions import RetryHttpRequestError, EndOfTriesError, UnexpectedError, ApiProxyError


class ApiProxyService:
    def __init__(self, app, default_headers):
        """
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__default_headers = default_headers
        self.__options = {}
        self.__data_get_cache = {}
        self.__data_get_flatten_cache = {}

    def call_proxy(self, engine, payload, method, analyze_json_error_param, retry_request_substr_variants, stream=False):
        """
        :param engine: Система
        :param payload: Данные для запроса
        :param method: string Может содержать native_call | tsv | json_newline
        :param analyze_json_error_param: Нужно ли производить анализ параметра error в ответе прокси
        :param retry_request_substr_variants: Список подстрок, при наличии которых в ответе будет происходить перезапрос
        :param stream:
        :return:
        """
        log_ctx = {
            "engine": engine,
            "method": payload.get('method'),
            "method_params": payload.get('method_params')
        }
        self.__app.log.info("Call api proxy", log_ctx)
        body = {
            "engine": engine,
            "payload": payload
        }
        for _try_idx in range(20):
            try:
                # 1h таймаут, так как бывают большие долгие данные, а лимит хоть какой-то нужен
                body_str = json.dumps(body)
                headers = {
                    "User-Agent": self.__app.user_agent,
                    "X-App": "META",
                    "X-Worker": self.__app.service_id,
                    "X-ObjectLocator": LOGGER_ENTITY.get("objectLocator")
                }
                resp = requests.post(self.__app.api_proxy_url + "/" + method, body_str, timeout=3600, stream=stream, headers=headers)

                self.check_err(resp, analyze_json_error_param=analyze_json_error_param, retry_request_substr_variants=retry_request_substr_variants)
                return resp
            except RetryHttpRequestError as e:
                self.__app.log.warning("Sleep retry query: " + str(e), log_ctx)
                time.sleep(20)
        raise EndOfTriesError("Api of api proxy tries request")

    @staticmethod
    def check_err(resp, analyze_json_error_param=False, retry_request_substr_variants=None):
        """
        :type retry_request_substr_variants: list Список вхождений строк, при налиции которых в ошидке апи будет произведен повторный запрос к апи
        """
        if retry_request_substr_variants is None:
            retry_request_substr_variants = []

        # РКН блокировки вызывают ошибку SSL
        retry_request_substr_variants.append("TLSV1_ALERT_ACCESS_DENIED")

        if resp.status_code in [502, 503, 504]:
            raise RetryHttpRequestError(resp.text)

        if resp.status_code >= 400:
            rtext = resp.text
            for v_ in retry_request_substr_variants:
                if v_ in rtext:
                    raise RetryHttpRequestError(rtext)
            raise UnexpectedError("HTTP request failed: {} {}".format(resp.status_code, rtext))
        if analyze_json_error_param:
            data_ = resp.json()
            if 'error' in data_ and data_.get('error'):
                full_err_ = json.dumps(data_.get('error'))
                for v_ in retry_request_substr_variants:
                    if v_ in full_err_:
                        raise RetryHttpRequestError(full_err_)
                raise ApiProxyError(full_err_)
        return resp



# Более подробное описание в документации по ошибкам что в каком случае использовать


class SDKError(Exception):
    """
    Корневая ошибка для SDK
    """
    pass


class BadRequestError(SDKError):
    """
    Запрос принципиально неправильный (неверная версия api, не передан обязательный параметр и пр.) (HTTP 400)
    """
    pass


class AuthError(SDKError):
    """
    Невозможно авторизоваться (HTTP 401)
    """
    pass


class ForbiddenError(SDKError):
    """
    Нет прав для соверщения операции (HTTP 403)
    """
    pass


class ServerError(SDKError):
    """
    Сервер ответил ошибкой (HTTP >500)
    """
    pass


class NoContentError(SDKError):
    """
    Нет содержимого (HTTP 204)
    """
    pass


class RequestError(SDKError):
    """
    Ошибка исполнения запроса(HTTP >400)
    """
    pass


class UnexpectedError(SDKError):
    """
    Непредвиденная ошибка (HTTP other)
    """
    pass


class DbQueryError(SDKError):
    """
    Ошибка работы с базой данных
    """
    pass


class RetryHttpRequestError(SDKError):
    """
    Невозможность повторного запроса (HTTP 502, 503, 504)
    """
    pass


class ApiProxyError(SDKError):
    """
    Ошибка работы с прокси
    """
    pass


class EndOfTriesError(SDKError):
    """
    Достигнут лимит количества повторов запроса
    """
    pass


class LockServiceError(SDKError):
    """
    Ошибка при работе с сервисом блокировки очереди (LockService)
    """
    pass



import json

import shutil


class MetaqlService:
    def __init__(self, app, default_headers):
        """
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__default_headers = default_headers
        self.__options = {}

    def download_data(self, configuration, output_file):
        """
        Выполняет указанный в конфигурации запрос и отдает файл на скачивание
        :param configuration: Конфгурация запроса
        :param output_file: Место, куда надо скачать файл
        :return:
        """
        params = configuration
        response = self.__app.native_api_call('metaql', 'download-data', params, self.__options, False, None, True, http_path="/api/v1/meta/")
        with open(output_file, 'wb') as out_file:
            shutil.copyfileobj(response.raw, out_file)
        del response

    def get_schema(self, db_alias, entity):
        """
        Возвращает схему сущности:
            - Поля
        :param db_alias: Альяс БД
        :param entity: Альяс Сущности
        :return: dict
        """
        response = self.__app.native_api_call('metaql', 'schema/' + db_alias + '/' + entity, {}, self.__options, False, None, False, http_path="/api/v1/meta/", http_method="GET")
        return json.loads(response.text)



import json
import os
import sys
import time

import requests

from metasdk.exceptions import UnexpectedError, DbQueryError, ServerError
from metasdk.internal import read_developer_settings
from metasdk.logger import create_logger, eprint
from metasdk.logger.bulk_logger import BulkLogger
from metasdk.logger.logger import Logger
from metasdk.services import get_api_call_headers, process_meta_api_error_code
from metasdk.services.ApiProxyService import ApiProxyService
from metasdk.services.DbQueryService import DbQueryService
from metasdk.services.DbService import DbService
from metasdk.services.ExportService import ExportService
from metasdk.services.ExternalSystemService import ExternalSystemService
from metasdk.services.FeedService import FeedService
from metasdk.services.IssueService import IssueService
from metasdk.services.MediaService import MediaService
from metasdk.services.MetaqlService import MetaqlService
from metasdk.services.SettingsService import SettingsService
from metasdk.services.UserManagementService import UserManagementService
from metasdk.services.StarterService import StarterService
from metasdk.services.MailService import MailService
from metasdk.services.LockService import LockService
from metasdk.worker import Worker


class MetaApp(object):
    debug = False
    service_id = None
    build_num = None
    starter_api_url = None
    meta_url = None
    api_proxy_url = None
    log = Logger()
    worker = None

    # Будет поставляться в конец UserAgent
    user_agent_postfix = ""

    developer_settings = None

    # Пользователь, из под которого пройдет авторизация после того,
    # как мета авторизует разработчика, в случае, если разработчик имеет разрешения для авторизации из-под других пользователей
    auth_user_id = None

    MediaService = None
    MetaqlService = None
    ExportService = None
    SettingsService = None
    IssueService = None
    UserManagementService = None
    StarterService = None
    MailService = None
    ApiProxyService = None

    __default_headers = set()
    __db_list = {}

    def __init__(self, service_id: str = None, debug: bool = None,
                 starter_api_url: str = None,
                 meta_url: str = None,
                 api_proxy_url: str = None,
                 include_worker: bool = None
                 ):
        if debug is None:
            is_prod = os.environ.get('PRODUCTION', False)
            debug = os.environ.get('DEBUG', not is_prod)
            if include_worker is None:
                include_worker = True
            if debug == 'false':
                debug = False
        self.debug = debug

        self.redis_url = os.environ.get("REDIS_URL", meta_url or "s1.meta.vmc.loc:6379:1")
        self.meta_url = os.environ.get("META_URL", meta_url or "http://apimeta.devision.io")
        self.api_proxy_url = os.environ.get("API_PROXY_URL", api_proxy_url or "http://apiproxy.apis.kb.1ad.ru")

        if debug and not starter_api_url:
            starter_api_url = "http://STUB_URL"
        self.starter_api_url = os.environ.get("STARTER_URL", starter_api_url or "http://s2.meta.vmc.loc:28341")

        if service_id:
            self.log.warning("Параметр service_id скоро будет удален из MetaApp")

        gcloud_log_host_port = os.environ.get("GCLOUD_LOG_HOST_PORT", "n3.adp.vmc.loc:31891")
        service_ns = os.environ.get('SERVICE_NAMESPACE', "appscript")  # для ns в логах
        service_id = os.environ.get('SERVICE_ID', "local_debug_serivce")
        self.build_num = os.environ.get('BUILD_NUM', '0')
        self.service_id = service_id
        create_logger(service_id=service_id, service_ns=service_ns, build_num=self.build_num, gcloud_log_host_port=gcloud_log_host_port, debug=self.debug)

        self.__read_developer_settings()

        self.__default_headers = get_api_call_headers(self)
        self.MediaService = MediaService(self, self.__default_headers)
        self.MetaqlService = MetaqlService(self, self.__default_headers)
        self.SettingsService = SettingsService(self, self.__default_headers)
        self.ExportService = ExportService(self, self.__default_headers)
        self.IssueService = IssueService(self, self.__default_headers)
        self.StarterService = StarterService(self, self.__default_headers)
        self.MailService = MailService(self, self.__default_headers)
        self.DbService = DbService(self, self.__default_headers)
        self.UserManagementService = UserManagementService(self, self.__default_headers)
        self.ApiProxyService = ApiProxyService(self, self.__default_headers)
        self.ExternalSystemService = ExternalSystemService(self, self.__default_headers)
        self.FeedService = FeedService(self, self.__default_headers)
        self.LockService = LockService(self)

        if include_worker:
            stdin = "[]" if debug else ''.join(sys.stdin.readlines())
            self.worker = Worker(self, stdin)

    def bulk_log(self, log_message=u"Еще одна пачка обработана", total=None, part_log_time_minutes=5):
        """
        Возвращает инстант логгера для обработки списокв данных
        :param log_message: То, что будет написано, когда время придет
        :param total: Общее кол-во объектов, если вы знаете его
        :param part_log_time_minutes: Раз в какое кол-во минут пытаться писать лог
        :return: BulkLogger
        """
        return BulkLogger(log=self.log, log_message=log_message, total=total, part_log_time_minutes=part_log_time_minutes)

    def db(self, db_alias, shard_key=None):
        """
        Получить экземпляр работы с БД
        :type db_alias: basestring Альяс БД из меты
        :type shard_key: Любой тип. Некоторый идентификатор, который поможет мете найти нужную шарду. Тип зависи от принимающей стороны
        :rtype: DbQueryService
        """
        if shard_key is None:
            shard_key = ''

        db_key = db_alias + '__' + str(shard_key)
        if db_key not in self.__db_list:
            self.__db_list[db_key] = DbQueryService(self, self.__default_headers, {"db_alias": db_alias, "dbAlias": db_alias, "shard_find_key": shard_key, "shardKey": shard_key})
        return self.__db_list[db_key]

    @property
    def user_agent(self):
        return self.service_id + " | b" + self.build_num + (' | ' + self.user_agent_postfix if self.user_agent_postfix else "")

    def __read_developer_settings(self):
        """
        Читает конфигурации разработчика с локальной машины или из переменных окружения
        При этом переменная окружения приоритетнее
        :return:
        """
        self.developer_settings = read_developer_settings()
        if not self.developer_settings:
            self.log.warning("НЕ УСТАНОВЛЕНЫ настройки разработчика, это может приводить к проблемам в дальнейшей работе!")

    def api_call(self, service, method, data, options):
        """
        :type app: metasdk.MetaApp
        """
        if 'self' in data:
            # может не быть, если вызывается напрямую из кода,
            # а не из прослоек типа DbQueryService
            data.pop("self")

        if options:
            data.update(options)

        _headers = dict(self.__default_headers)

        if self.auth_user_id:
            _headers['X-META-AuthUserID'] = str(self.auth_user_id)

        request = {
            "url": self.meta_url + "/api/v1/adptools/" + service + "/" + method,
            "data": json.dumps(data),
            "headers": _headers,
            "timeout": (60, 1800)
        }

        for _try_idx in range(20):
            try:
                resp = requests.post(**request)
                if resp.status_code == 200:
                    decoded_resp = json.loads(resp.text)
                    if 'data' in decoded_resp:
                        return decoded_resp['data'][method]
                    if 'error' in decoded_resp:
                        if 'details' in decoded_resp['error']:
                            eprint(decoded_resp['error']['details'])
                        raise DbQueryError(decoded_resp['error'])
                    raise UnexpectedError()
                else:
                    process_meta_api_error_code(resp.status_code, request, resp.text)
            except (requests.exceptions.ConnectionError, ConnectionError, TimeoutError) as e:
                self.log.warning('META API Connection Error. Sleep...', {"e": e})
                time.sleep(15)
            except Exception as e:
                if 'Служба частично или полностью недоступна' in str(e):
                    self.log.warning('META API Connection Error. Sleep...', {"e": e})
                    time.sleep(15)
                else:
                    raise e

        raise ServerError(request)

    def native_api_call(self, service, method, data, options, multipart_form=False, multipart_form_data=None, stream=False, http_path="/api/meta/v1/", http_method='POST',
                        get_params=None, connect_timeout_sec=60):
        """
        :type app: metasdk.MetaApp
        :rtype: requests.Response
        """
        if get_params is None:
            get_params = {}
        if 'self' in data:
            # может не быть, если вызывается напрямую из кода,
            # а не из прослоек типа DbQueryService
            data.pop("self")

        if options:
            data.update(options)

        _headers = dict(self.__default_headers)

        if self.auth_user_id:
            _headers['X-META-AuthUserID'] = str(self.auth_user_id)

        request = {
            "url": self.meta_url + http_path + service + "/" + method,
            "timeout": (connect_timeout_sec, 1800),
            "stream": stream,
            "params": get_params,
        }

        if multipart_form:
            if multipart_form_data:
                request['files'] = multipart_form_data
            request['data'] = data
            _headers.pop('content-type', None)
        else:
            request['data'] = json.dumps(data)
        request['headers'] = _headers

        for _try_idx in range(20):
            try:
                resp = requests.request(http_method, **request)
                if resp.status_code == 200:
                    return resp
                else:
                    process_meta_api_error_code(resp.status_code, request, resp.text)
            except (requests.exceptions.ConnectionError, ConnectionError, TimeoutError) as e:
                self.log.warning('META API Connection Error. Sleep...', {"e": e})
                time.sleep(15)
            except Exception as e:
                if 'Служба частично или полностью недоступна' in str(e):
                    self.log.warning('META API Service Temporarily Unavailable. Sleep...', {"e": e})
                    time.sleep(15)
                else:
                    raise e

        raise ServerError(request)

    def get_lib_version(self):
        from metasdk import info
        return info.__version__



"""

Супер мелкие функции, которые нужны от 3 использований

"""
import json
from itertools import islice
import jwt


def chunks_generator(iterable, count_items_in_chunk):
    """
    Очень внимательно! Не дает обходить дважды

    :param iterable:
    :param count_items_in_chunk:
    :return:
    """
    iterator = iter(iterable)
    for first in iterator:  # stops when iterator is depleted
        def chunk():  # construct generator for next chunk
            yield first  # yield element from for loop
            for more in islice(iterator, count_items_in_chunk - 1):
                yield more  # yield more elements from the iterator

        yield chunk()  # in outer generator, yield next chunk


def chunks(list_, count_items_in_chunk):
    """
    разбить list (l) на куски по n элементов

    :param list_:
    :param count_items_in_chunk:
    :return:
    """
    for i in range(0, len(list_), count_items_in_chunk):
        yield list_[i:i + count_items_in_chunk]


def pretty_json(obj):
    """
    Представить объект в вище json красиво отформатированной строки
    :param obj:
    :return:
    """
    return json.dumps(obj, sort_keys=True, indent=4, separators=(',', ': '), ensure_ascii=False)


def decode_jwt(input_text, secure_key):
    """
    Раскодирование строки на основе ключа
    :param input_text: исходная строка
    :param secure_key: секретный ключ
    :return:
    """
    if input_text is None:
        return None

    encoded = (input_text.split(":")[1]).encode('utf-8')
    decoded = jwt.decode(encoded, secure_key)
    return decoded['sub']



import requests

import json
import time
import base64

from metasdk import read_developer_settings
from metasdk.exceptions import SDKError
from metasdk.internal import read_cfg, write_cfg

AUTH_CACHE_FILE = "/.rwmeta_auth_cache.json"
OAUTH_TOKEN_URL = "https://account.devision.io/oauth2/token"


class ApiClient:
    """
    Упрощает взаимодействие с внутренними API Devision
    """

    def __init__(self, host, api_version, access_token=None, refresh_token=None, client_id=None, client_secret=None):
        """
        :type app: metasdk.MetaApp
        """
        self.host = host
        self.api_version = api_version
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.client_id = client_id
        self.client_secret = client_secret

    def get(self, method_path, get_params=None):
        return self.request("GET", method_path, get_params)

    def post(self, method_path, get_params=None, post_data=None):
        return self.request("POST", method_path, get_params, post_data)

    def request(self, http_method, method_path, get_params=None, post_data=None):
        auth_cache = read_cfg(AUTH_CACHE_FILE)
        if auth_cache:
            if self.refresh_token == auth_cache.get('refresh_token', ''):
                self.access_token = self.__fetch_and_validate_access_token(auth_cache)

        if not self.access_token:
            self.refresh_access_token()

        if not self.access_token:
            raise ValueError("access_token должен был быть получен")

        for _try_idx in range(2):
            url = self.host + '/' + self.api_version + '/' + method_path
            headers = {"Authorization": "Bearer " + self.access_token}

            req_param = {
                "method": http_method,
                "url": url,
                "data": get_params,
                "json": post_data,
                "headers": headers
            }
            resp = requests.request(**req_param)
            if resp.status_code == 401:
                self.refresh_access_token()
                continue

            return resp.json()

    def refresh_access_token(self):
        resp = requests.post(OAUTH_TOKEN_URL, data={
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "refresh_token": self.refresh_token,
            "grant_type": "refresh_token",
        }, headers={
            "Content-Type": "application/x-www-form-urlencoded"
        })

        json = resp.json()
        if json.get('error'):
            raise SDKError(json)
        self.access_token = json.get('access_token')

        write_cfg(AUTH_CACHE_FILE, {
            "refresh_token": self.refresh_token,
            "access_token": self.access_token
        })

    @staticmethod
    def build_from_developer_settings(api_name: str, api_version: str):
        """
        :param api_name: Example hello
        :param api_version: Example v1, v2alpha
        :return: ApiClient
        """
        developer_settings = read_developer_settings()

        api_host = "http://" + api_name + ".apis.devision.io"
        return ApiClient(
            host=api_host,
            api_version=api_version,
            access_token=None,
            refresh_token=developer_settings['refreshToken'],
            client_id=developer_settings['clientId'],
            client_secret=developer_settings['clientSecret'],
        )

    @staticmethod
    def __fetch_and_validate_access_token(auth_cache):
        access_token = auth_cache.get('access_token', '')
        t = json.loads(base64.b64decode(access_token.split(".")[1] + "=").decode(encoding="utf-8"))
        if time.time() > t.get('exp'):
            return None
        return access_token



#!/usr/bin/env python
from os import path

import setuptools


def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]


from metasdk import info

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst')) as f:
    long_description = f.read()

packages = [
    'metasdk',
    'metasdk.logger',
    'metasdk.services'
]

install_reqs = parse_requirements('requirements.txt')
reqs = install_reqs

setuptools.setup(
    name=info.__package_name__,
    version=info.__version__,

    description='Devision Meta SDK',
    long_description=long_description,

    url='https://github.com/devision-io/metasdk',

    author='Artur Geraschenko',
    author_email='arturgspb@gmail.com',

    license='MIT',

    classifiers=[
        'Programming Language :: Python :: 3'
    ],
    install_requires=reqs,
    packages=packages,
    package_data={'': ['LICENSE']},
    package_dir={'metasdk': 'metasdk'},
    include_package_data=True,
)



import random
import time
import redis
from contextlib import contextmanager

from metasdk.exceptions import LockServiceError


class LockService:
    def __init__(self, app):
        """
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__redis_storage = None
        self.__unique_name = None
        self.__key = None
        self.__names = None

    def __connect_to_redis(self):
        if not self.__redis_storage:
            url_parts = self.__app.redis_url.split(':')
            self.__redis_storage = redis.Redis(host=url_parts[0], port=url_parts[1], db=url_parts[2])

    @contextmanager
    def lock(self, key: str, ttl_in_sec: int, timeout_in_sec: int, queue_width: int = 1):
        is_set = None
        self.__connect_to_redis()
        waiting_time = random.randint(1, 3)
        time.sleep(0.0001 * waiting_time)
        key = key + str(random.randint(1, queue_width))

        try:
            while timeout_in_sec > 0:
                is_set = self.__redis_storage.set(name=key, value=True, ex=ttl_in_sec, nx=True)
                if is_set:
                    yield
                    break
                else:
                    time.sleep(waiting_time)
                    timeout_in_sec -= waiting_time
            else:
                # Этот "else" относится к "while" и выполнится если не отработает "break"
                msg = "Функция с ключом {}, не смогла захватить лок за отведенное время".format(key)
                self.__app.log.warning(msg)
                raise LockServiceError(msg)

        except redis.exceptions.ConnectionError as e:
            error_msg = ("Ошибка подключения к redis", {"e": e})
            self.__app.log.error(*error_msg)
            raise LockServiceError(*error_msg)
        finally:
            if is_set:
                self.__redis_storage.delete(key)

    def lock_decorator(self, key: str, ttl_in_sec: int, timeout_in_sec: int, queue_width: int = 1, field_to_uniq=None):
        self.__key = key

        def decorator(func):
            def wrapper(*args, **kwargs):
                self.__unique_name = self.__key

                if field_to_uniq and kwargs:
                    for key, value in kwargs.items():
                        if key in field_to_uniq:
                            self.__unique_name += '_{0}-{1}'.format(key, value)

                with self.lock(key=self.__unique_name, ttl_in_sec=ttl_in_sec, timeout_in_sec=timeout_in_sec, queue_width=queue_width):
                    return func(*args, **kwargs)

            return wrapper
        return decorator

